# -*- coding: utf-8 -*-
"""
Define constants and function dedicated to RF-conception
"""
import numpy as np
from numba import njit, vectorize

# Constants
u0 = 4 * np.pi * 1e-7  # H/m
c0 = 299792458  # m/s
eps0 = 8.8541878128e-12  # F/m
eta0 = np.sqrt(u0 / eps0)  # Ohm
Nm_to_dBcm = 8.686 / 100


# Other functions
@vectorize(cache=True)
def gamma(_z_load, _z0=50):
    return (_z0 - _z_load) / (_z0 + _z_load)


@njit(cache=True)
def std_dev(measured, targeted):
    """
    return the standard deviation between an array_like of results and their references.
    """
    tmp = np.abs(gamma(measured, targeted)) ** 2
    return np.sqrt(np.sum(tmp))


@vectorize(cache=True)
def dB(c):
    """
    Return the decibel value of the given imaginary number.
    """
    if c == 0:
        return -np.inf
    return 10 * np.log10(np.abs(c))


@vectorize(cache=True)
def lin(d):
    """
    return the linear magnitude of the given magnitude in decibel
    """
    return 10 ** (d / 10)


@vectorize(cache=True)
def ihsr(_s31, _s21):
    """
    Return the IHSR (Ideal Hybrid Splitting Ratio) for the given gains
    """
    if _s21 == 1j * _s31 or _s21 == -1j * _s31:
        return np.inf
    ihsr_log = dB((_s21 - 1j * _s31) / (_s21 + 1j * _s31))
    return -np.minimum(ihsr_log, -ihsr_log)


@vectorize(cache=True)
def quality_f(_z):
    """
    return the quality factor of a components
    """
    return np.imag(_z) / np.real(_z)


@njit(cache=True)
def reflexion_coef(_z_steps, _phi_step):
    """
    return the coefficient reflexion of a given sequence of transmission lines
    with the given_z_steps profile and equal length of _phi_steps degrees
    """
    n_step = len(_z_steps)

    z_tot = _z_steps[-1:]
    for i in range(n_step):
        z_0 = _z_steps[n_step - i - 1]
        z_tot = (
            z_0
            * (z_tot + 1j * z_0 * np.tan(_phi_step))
            / (z_0 + 1j * z_tot * np.tan(_phi_step))
        )
    return gamma(_z_steps[0], z_tot)


@njit(cache=True)
def transmission_coef(_z_steps, _phi_step):
    return np.sqrt(1 - reflexion_coef(_z_steps, _phi_step) ** 2)


@vectorize(cache=True)
def frac_bandwidth(f_min, f_max):
    return 100 * np.abs(f_max - f_min) / np.sqrt(f_max * f_min)


@njit(cache=True)
def friis(f, gain):
    """

    Parameters
    ----------
    f : np.Array
        List of the noise figure (in dB) of each block.
    gain : np.Array
        List of the gain of each block (in dB).


    Returns
    -------
    float
        Total noise figure of the system

    """

    m = gain.shape[0]
    n = f.shape[0]
    if m != n - 1:
        raise ValueError("gain should have 1 item less than noise factor f")
    g_tot = 1
    f_lin = lin(f)
    res = f_lin[0]
    for i in range(m):
        g_tot *= lin(gain[i])
        res += (f_lin[i + 1] - 1) / g_tot
    return dB(res)


def int_phase_noise(pn_db, freq, f_min=None, f_max=None):
    """

    Parameters
    ----------
    pn_db : np.Array
        List of the phase noise in dBc/Hz
    freq : np.Array
        List of the corresponding frequency (in Hz)
    f_min, f_max : float
        if not None, calculation is done from f_min to f_max (with interpolation)


    Returns
    -------
    float
        integrated phase noise of the piece wise phase noise curve (in radian)

    """
    pn_shape = pn_db.size
    f_shape = freq.size
    if pn_shape != f_shape:
        raise ValueError(f"Expected identical shape, got {pn_shape} and {f_shape}")
    ipn = 0
    for i in range(pn_shape - 1):
        if f_min is not None and f_min > freq[i + 1]:
            # skipping all part bellow f_min
            continue
        if f_min is not None and f_min > freq[i]:
            f1 = f_min
            pn1_db = __pn_interpol(pn_db[i : i + 2], freq[i : i + 2], f_min)
        else:
            f1 = freq[i]
            pn1_db = pn_db[i]

        if f_max is not None and freq[i] > f_max:
            continue
        if f_max is not None and freq[i + 1] < f_max:
            f2 = f_max
            pn2_db = __pn_interpol(pn_db[i : i + 2], freq[i : i + 2], f_max)
        else:
            f2 = freq[i + 1]
            pn2_db = pn_db[i + 1]
        A = (pn1_db - pn2_db) / dB(f1 / f2)
        ipn += (f2 * lin(pn2_db) - f1 * lin(pn1_db)) / (A + 1)
    return ipn


def __pn_interpol(pn_db, freq, f_int):
    """
    interpolator for phase noise
    """
    a = (pn_db[1] - pn_db[0]) / (10 * np.log10(freq[1] / freq[0]))
    b = pn_db[1] - a * 10 * np.log10(freq[1])
    return b + a * 10 * np.log10(f_int)


def ipn_to_jitter(ipn, f0):
    """

    Parameters
    ----------
    ipn : float
        integrated phase noise of an system
    f0 : float
        central frequency of the system

    Returns
    -------
    float
        equivalent jitter in seconds

    """
    return np.sqrt(2 * ipn) / (2 * np.pi * f0)



# -*- coding: utf-8 -*-
"""

"""
import numpy as np
import yaml
import skrf as rf
from passive_auto_design.special import u0
import passive_auto_design.components.lumped_element as lmp


class Transformer:
    """
    Create a transformer object with the specified geometry _primary & _secondary
    (which are dict defined as :
        {'di':_di,'n_turn':_n_turn, 'width':_width, 'gap':_gap, 'height':height})
    and calculate the associated electrical model
    """

    def __init__(self, primary, secondary, freq=1e9, model_map_file=None):
        self.prim = primary
        self.second = secondary
        if isinstance(freq, rf.Frequency):
            self.freq = freq
        elif isinstance(freq, float):
            self.freq = rf.Frequency(freq, freq, 1, unit="Hz")
        else:
            self.freq = rf.Frequency.from_f(freq, unit="Hz")
        if model_map_file is None:
            model_map_file = "tests/default.map"
        with open(model_map_file, "r") as file:
            self.model_map = yaml.full_load(file)
        self.model = {}
        self.set_primary(primary)
        self.set_secondary(secondary)

    def set_primary(self, _primary):
        """
        modify the top inductor and refresh the related model parameters
        """
        self.prim = _primary
        self.model.update(
            {
                "lp": self.l_geo(True),
                "rp": self.r_geo(True),
                "cg": self.cc_geo(False),
                "cm": self.cc_geo(True),
                "k": self.k_geo(),
            }
        )
        try:
            self.circuit = self.__make_circuit()
        except KeyError:
            pass

    def set_secondary(self, _secondary):
        """
        modify the bottom inductor and refresh the related model parameters
        """
        self.second = _secondary
        self.model.update(
            {
                "ls": self.l_geo(False),
                "rs": self.r_geo(False),
                "cg": self.cc_geo(False),
                "cm": self.cc_geo(True),
                "k": self.k_geo(),
            }
        )
        try:
            self.circuit = self.__make_circuit()
        except KeyError:
            pass

    def l_geo(self, _of_primary=True):
        """
        return the value of the distributed inductance of the described transformer
        if _of_primary, return the value of the top inductor
        else, return the value of the bottom inductor
        """
        k_1 = float(self.model_map["mu_r"])  # constante1 empirique pour inductance
        k_2 = float(self.model_map["dens"])  # constante2 empirique pour inductance
        if _of_primary:
            geo = self.prim
        else:
            geo = self.second
        outer_diam = (
            geo["di"]
            + 2 * geo["n_turn"] * geo["width"]
            + 2 * (geo["n_turn"] - 1) * geo["gap"]
        )
        rho = (geo["di"] + outer_diam) / 2
        density = (outer_diam - geo["di"]) / (outer_diam + geo["di"])
        return k_1 * u0 * geo["n_turn"] ** 2 * rho / (1 + k_2 * density)

    def cc_geo(self, _mutual=True):
        """
        return the value of the distributed capacitance of the described transformer
        if _mutual, return the capacitance between primary and secondary
        else, return the capacitance to the ground plane
        """
        if _mutual:
            dist = float(self.model_map["d_m"])
        else:
            dist = float(self.model_map["d_g"])
        n_t = self.prim["n_turn"]
        eps_r = float(self.model_map["eps_r"])
        area = n_t * self.prim["di"] * self.prim["width"]
        cap = lmp.Capacitor(self.freq, area, dist, eps_r)
        return cap.par["cap"]

    def r_geo(self, _of_primary=True):
        """
        return the value of the resistance of the described transformer
        """
        if _of_primary:
            geo = self.prim
        else:
            geo = self.second
        rho = self.model_map["rho"]
        n_t = geo["n_turn"]
        l_tot = (
            8
            * np.tan(np.pi / 8)
            * n_t
            * (geo["di"] + geo["width"] + (n_t - 1) * (geo["width"] + geo["gap"]))
        )
        r_dc = rho * l_tot / geo["width"]
        return np.maximum(r_dc, 0)

    def k_geo(self):
        """
        return the value of the coupling between the two inductors.

        """
        return self.model_map["cpl_eq"]

    def mutual_geo(self, z_0=50):
        """
        Create a transformateur with a primary of l_1, and secondary of l_2
        and a coupling factor of k_mut
        """
        l_1 = self.model["lp"]
        l_2 = self.model["ls"]
        r_1 = self.model["rp"]
        r_2 = self.model["rs"]
        k_mut = self.model["k"]
        for f_t in self.freq.f:
            w_t = float(2 * np.pi * f_t)
            y_1 = 1 / (r_1 + 1j * w_t * l_1 * (1 - k_mut ** 2))
            y_2 = 1 / (r_2 + 1j * w_t * l_2 * (1 - k_mut ** 2))
            y_m = k_mut / (1e-30 + 1j * w_t * np.sqrt(l_1 * l_2) * (1 - k_mut ** 2))
            y_param = np.array(
                [
                    [
                        [y_1, -y_m, y_m, -y_1],
                        [-y_m, y_2, -y_2, y_m],
                        [y_m, -y_2, y_2, -y_m],
                        [-y_1, y_m, -y_m, y_1],
                    ]
                ]
            )
            try:
                y_params = np.vstack((y_params, y_param))
            except NameError:  # Only needed for first iteration.
                y_params = np.copy(y_param)
        ntwk = rf.Network(
            frequency=self.freq, s=rf.y2s(y_params), z0=z_0, name="coupled inductors"
        )
        return ntwk

    def __make_circuit(self):
        media = rf.media.DefinedGammaZ0(frequency=self.freq, Z0=50)
        transfo_ideal = self.mutual_geo()
        cap_g, ports = [], []
        for i in range(4):
            cap_g.append(media.capacitor(self.model["cg"], name=f"cg{i}"))
            ports.append(rf.Circuit.Port(self.freq, f"port{i}"))
        cap_m1 = media.capacitor(self.model["cm"], name="cm1")
        cap_m2 = media.capacitor(self.model["cm"], name="cm2")

        connections = []
        for i in range(4):
            connections.append([(ports[i], 0), (transfo_ideal, i), (cap_g[i], 0)])
        connections.append([(cap_m1, 0), (transfo_ideal, 0)])
        connections.append([(cap_m1, 1), (transfo_ideal, 1)])
        connections.append([(cap_m2, 0), (transfo_ideal, 3)])
        connections.append([(cap_m2, 1), (transfo_ideal, 2)])
        cir = rf.Circuit(connections)
        return cir



# -*- coding: utf-8 -*-
"""

"""
import abc
import functools
from scipy.optimize import minimize_scalar
import skrf as rf
from passive_auto_design.special import u0, eps0


class LumpedElement(metaclass=abc.ABCMeta):
    """
    class of standard lumped element, to be inherited by all lumped elements
    """

    def __init__(self, freq, z0):
        self._par = {}
        self.ref = None
        self.line = rf.media.DefinedGammaZ0(freq, z0=z0)

    @property
    def par(self):
        """
        Returns
        -------
        Dictionary
            Parameters of the lumped component.

        """
        return self._par

    @par.setter
    def par(self, key_pair):
        self._par.update(key_pair)

    def set_x_with_y(self, x_key, y_key, y_value):
        """
        set the value of the X_key for a Y_value of the Y_key
        """
        minimize_scalar(self.__cost, args=(x_key, y_key, y_value))

    @functools.lru_cache()
    def __cost(self, x_value, x_key, y_key, y_value):
        self.par = {x_key: x_value}
        if isinstance(y_value, float):
            self.par = {y_key: y_value}
            return abs(self.calc_ref_value() - self.par[self.ref])
        err = 0
        for y_val in y_value:
            self.par = {y_key: y_val}
            err += abs(self.calc_ref_value() - self.par[self.ref])
        return err

    @abc.abstractmethod
    def calc_ref_value(self):
        """
        Definition of the behavioral equation of the lumped element.
        Returns
        -------
        Value of self.par[self.ref] calculated with the other value in self.par.

        """
        raise NotImplementedError


class Resistor(LumpedElement):
    """
    class describing a resistor behavior
    """

    def __init__(self, freq, section=1e-3, length=1, rho=1e-15, z0=50):
        LumpedElement.__init__(self, freq, z0)
        self.par = {
            "section": section,
            "length": length,
            "rho": rho,
        }
        self.ref = "res"
        self.par.update({"res": self.calc_ref_value()})
        self.sp = self.line.resistor(self.par[self.ref])

    def calc_ref_value(self):
        return self.par["rho"] * self.par["length"] / self.par["section"]


class Capacitor(LumpedElement):
    """
    class describing a capacitor behavior
    """

    def __init__(self, freq, area=1e-6, dist=1e-3, eps_r=1, z0=50):
        LumpedElement.__init__(self, freq, z0)
        self.ref = "cap"
        self.par = {
            "eps_r": eps_r,
            "area": area,
            "dist": dist,
        }
        self.par.update({"cap": self.calc_ref_value()})
        self.sp = self.line.capacitor(self.par[self.ref])

    def calc_ref_value(self):
        return eps0 * self.par["eps_r"] * self.par["area"] / self.par["dist"]


class Inductor(LumpedElement):
    """
    class describing a inductor behavior
    """

    def __init__(self, freq, d_i=100e-6, n_turn=1, width=3e-6, gap=1e-6, z0=50):
        LumpedElement.__init__(self, freq, z0)
        self.ref = "ind"
        self.par = {
            "d_i": d_i,
            "n_turn": n_turn,
            "width": width,
            "gap": gap,
            "k_1": 1.265,
            "k_2": 2.093,
        }
        self.par.update({"ind": self.calc_ref_value()})
        self.sp = self.line.inductor(self.par[self.ref])

    def calc_ref_value(self):
        outer_diam = (
            self.par["d_i"]
            + 2 * self.par["n_turn"] * self.par["width"]
            + 2 * (self.par["n_turn"] - 1) * self.par["gap"]
        )
        self.par.update({"d_o": outer_diam})
        rho = (self.par["d_i"] + outer_diam) / 2
        density = (outer_diam - self.par["d_i"]) / (outer_diam + self.par["d_i"])
        return (
            self.par["k_1"]
            * u0
            * self.par["n_turn"] ** 2
            * rho
            / (1 + self.par["k_2"] * density)
        )


class Mutual(LumpedElement):
    """
    class describing a mutual inductor behavior
    """

    def __init__(self, freq, ind1, ind2, z0=50):
        self.ref = "k"
        LumpedElement.__init__(self, freq, z0)
        self.ind1 = ind1
        self.ind2 = ind2
        self.par = {"cpl_eq": 1}
        self.par.update({"k": self.calc_ref_value()})

    def calc_ref_value(self):
        cpl = self.par["cpl_eq"]
        return cpl * (
            min(self.ind1.par["d_o"], self.ind2.par["d_o"])
            - max(self.ind1.par["d_i"], self.ind2.par["d_i"])
        )



# -*- coding: utf-8 -*-
"""
This module give function to ease the design of RF-tapper
"""
import numpy as np
from scipy.special import i1
from scipy.integrate import quad
from ..special import gamma


def linear_taper(_z_start, _z_stop, _n_step):
    """
    return the _n_step profile of impedance for a transition from
    _z_start to _z_stop
    """
    return np.linspace(_z_start, _z_stop, _n_step, dtype=complex)


def klopfenstein_taper(_z_start, _z_stop, _n_step, _rhomax=0.01):
    """
    return the _n_step profile of impedance for a transition from
    _z_start to _z_stop
    """
    rho0 = gamma(_z_start, _z_stop)
    z_mid = 0.5 * np.log(_z_stop * _z_start)
    n_mid = int(np.floor(_n_step / 2))
    a_coeff = np.arccosh(rho0 / _rhomax)
    ln_z = np.zeros((_n_step,))
    for i in range(1, n_mid + 1):
        ln_z[i + n_mid] = z_mid + _rhomax * (
            1 + a_coeff ** 2 * __phi(a_coeff, i / n_mid)
        )
        ln_z[n_mid - i] = z_mid + _rhomax * (
            1 - a_coeff ** 2 * __phi(a_coeff, i / n_mid)
        )
    ln_z[n_mid] = z_mid + _rhomax
    return np.exp(ln_z, dtype=complex)


def __phi(_a_coeff, _y_pos):
    phi_r = quad(__phase_equation, 0, _y_pos, args=_a_coeff)
    return phi_r[0]


def __phase_equation(x, a):
    a = np.abs(a)
    return i1(a * np.sqrt(1 - x ** 2)) / (a * np.sqrt(1 - x ** 2))
